package org.jvnet.jaxbcommons.extendedi18n.addon;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.net.URI;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.TreeMap;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jvnet.jaxbcommons.addon.AbstractParameterizableCodeAugmenter;
import org.jvnet.jaxbcommons.enums.addon.util.EnumUtils;
import org.jvnet.jaxbcommons.i18n.addon.util.PropertyUtils;
import org.jvnet.jaxbcommons.util.ClassUtils;
import org.jvnet.jaxbcommons.util.FieldUtils;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;

import com.sun.codemodel.JDefinedClass;
import com.sun.msv.datatype.xsd.XSDatatype;
import com.sun.msv.grammar.ValueExp;
import com.sun.tools.xjc.Options;
import com.sun.tools.xjc.generator.ClassContext;
import com.sun.tools.xjc.generator.GeneratorContext;
import com.sun.tools.xjc.generator.PackageContext;
import com.sun.tools.xjc.grammar.AnnotatedGrammar;
import com.sun.tools.xjc.grammar.ClassItem;
import com.sun.tools.xjc.grammar.FieldItem;
import com.sun.tools.xjc.grammar.FieldUse;

public class AddOn extends AbstractParameterizableCodeAugmenter {

  /** Logger. */
  protected Log logger = LogFactory.getLog(getClass());

  /** Option name. */
  public static final String OPTION_NAME = "XextendedI18n";

  // Configurable properties

  private List locales = new ArrayList(3);

  public void setLocale(String locale) {
    locales.add(parseLocale(locale));
  }

  public Locale[] getLocales() {
    return (Locale[]) locales.toArray(new Locale[locales.size()]);
  }

  private File importDirectory;

  public void setImportDirectory(File importDirectory) {
    this.importDirectory = importDirectory;
  }

  private String importName = "messages";

  public void setImportName(String importName) {
    this.importName = importName;
  }

  private File exportDirectory;

  public void setExportDirectory(File exportDirectory) {
    this.exportDirectory = exportDirectory;
  }

  private String exportName = "messages";

  public void setExportName(String exportName) {
    this.exportName = exportName;
  }

  public String getOptionName() {
    return OPTION_NAME;
  }

  public String getUsage() {
    return "";
  }

  /**
   * Parses locale code given as string.
   * 
   * @param string
   *            locale code.
   * @return Parsed locale.
   */
  private Locale parseLocale(String string) {
    final StringTokenizer tokens = new StringTokenizer(string, "_");
    final String language = tokens.hasMoreTokens() ? tokens.nextToken() : "";
    final String country = tokens.hasMoreTokens() ? tokens.nextToken() : "";
    String variant = "";
    String sep = "";
    while (tokens.hasMoreTokens()) {
      variant += sep + tokens.nextToken();
      sep = "_";
    }
    return new Locale(language, country, variant);
  }

  public boolean run(
      AnnotatedGrammar grammar,
      GeneratorContext context,
      Options options,
      ErrorHandler errorHandler) {

    if (importDirectory == null) {
      logger.debug("Import directory is not specified, trying to guess it.");

      final InputSource[] grammars = options.getGrammars();
      for (int index = 0; index < grammars.length; index++) {
        final InputSource source = grammars[index];
        try {
          final String systemId = source.getSystemId();
          logger.debug("Trying out [" + systemId + "].");
          final URI uri = new URI(systemId);
          final File file = new File(uri);
          importDirectory = file.getParentFile();
        }
        catch (Exception ex) {
          // Ignore the exception
        }
      }
    }
    if (importDirectory == null) {
      logger.warn("Could not find out the import directory.");
    }

    if (exportDirectory == null) {
      logger.info("Export directory is not specified. Using XJC target directory.");
      exportDirectory = options.targetDir;
    }

    importProperties(context);

    processProperties(context);

    writeProperties(context);

    return true;
  }


  public void addClassEntry(ClassContext classContext) {
    final PackageContext packageContext = classContext._package;

    final Locale[] locales = getLocales();

    for (int index = 0; index < locales.length; index++) {
      final Locale locale = locales[index];
      final String property = getImportedClassProperty(classContext, locale);
      final String key = getClassKey(classContext);
      getProperties(packageContext, locale).put(
          key,
          property != null ? property : "@@TODO." + classContext.target.name + "@@");
    }
  }

  public void addFieldEntry(ClassContext classContext, FieldItem fieldItem) {
    final PackageContext packageContext = classContext._package;

    final Locale[] locales = getLocales();

    for (int index = 0; index < locales.length; index++) {
      final Locale locale = locales[index];
      final String property = getImportedFieldProperty(classContext, fieldItem, locale);
      final String key = getFieldKey(classContext, fieldItem);
      getProperties(packageContext, locale).put(
          key,
          property != null ? property : "@@TODO." + fieldItem.name + "@@");
    }
  }

  public void addEnumEntry(ClassContext classContext, FieldItem fieldItem) {

    final FieldUse fieldUse = FieldUtils.getFieldUse(classContext, fieldItem);
    final JDefinedClass enumClass = (JDefinedClass) fieldUse.type;

    final PackageContext packageContext = classContext.parent.getPackageContext(enumClass
        ._package());

    final Collection enumValues = EnumUtils.getValues(classContext, fieldItem);

    if (enumValues != null) {
      for (final Iterator iterator = enumValues.iterator(); iterator.hasNext();) {
        final ValueExp enumValue = (ValueExp) iterator.next();

        final String lexical;
        if (enumValue.dt instanceof XSDatatype)
          lexical = ((XSDatatype) enumValue.dt).convertToLexicalValue(enumValue.value, null);
        else
          lexical = enumValue.value.toString();

        final Locale[] locales = getLocales();

        final String key = getEnumKey(enumClass, lexical);
        for (int index = 0; index < locales.length; index++) {
          final Locale locale = locales[index];
          final String property = getImportedProperty(packageContext, locale, key);
          getProperties(packageContext, locale).put(
              key,
              property != null ? property : "@@TODO." + lexical + "@@");
        }
      }
    }
  }

  private String getEnumKey(final JDefinedClass enumClass, final String lexical) {
    return enumClass.fullName() + "." + lexical;
  }

  public String getClassKey(ClassContext classContext) {
    return ClassUtils.getFQCN(classContext, false);
  }

  public String getFieldKey(ClassContext classContext, FieldItem fieldItem) {
    return FieldUtils.getFQFN(classContext, fieldItem, false);
  }

  private String getImportedClassProperty(ClassContext classContext, Locale locale) {
    final PackageContext packageContext = classContext._package;
    final String key = getClassKey(classContext);

    final String property = getImportedProperty(packageContext, locale, key);

    return property;
  }

  private String getImportedFieldProperty(
      ClassContext classContext,
      FieldItem fieldItem,
      Locale locale) {
    final PackageContext packageContext = classContext._package;
    final String key = getFieldKey(classContext, fieldItem);

    final String property = getImportedProperty(packageContext, locale, key);

    if (property != null) {
      return property;
    }
    else {
      final ClassItem classItem = classContext.target;
      if (classItem.getSuperClass() == null) {
        return null;
      }
      else {
        final ClassContext superClassContext = classContext.parent.getClassContext(classItem.getSuperClass());

        final String superProperty = getImportedFieldProperty(superClassContext, fieldItem, locale);
        return superProperty;
      }
    }
  }

  public String getImportedProperty(PackageContext packageContext, Locale locale, String key) {
    final Map localeProperties = (Map) importedProperties.get(locale);

    if (localeProperties == null) {
      return null;
    }
    else {
      return (String) localeProperties.get(key);
    }
  }

  private Map importedProperties = Collections.EMPTY_MAP;

  public void importProperties(final GeneratorContext generatorContext) {

    final Locale[] locales = getLocales();
    importedProperties = new HashMap(locales.length);

    for (int index = 0; index < locales.length; index++) {
      final Locale locale = locales[index];

      final Map localeProperties = new HashMap();
      importedProperties.put(locale, localeProperties);
      final PackageContext[] packageContexts = getPackageContexts(generatorContext);

      for (int jndex = 0; jndex < packageContexts.length; jndex++) {
        final PackageContext packageContext = packageContexts[jndex];
        final Map properties = importProperties(packageContext, locale);
        localeProperties.putAll(properties);
      }
    }
  }


  public Map importProperties(PackageContext packageContext, Locale locale) {
    if (importDirectory != null) {
      final String draftPackageName = packageContext._package.name();
      final String packageName = draftPackageName == null ? "" : draftPackageName;

      final String packageDirectoryName = packageName.replace('.', File.separatorChar);

      final File packageDirectory = new File(importDirectory, packageDirectoryName);

      final String fileName = importName + "_" + locale.toString() + ".properties";

      final File file = new File(packageDirectory, fileName);

      if (file.isFile()) {

        logger.info(MessageFormat.format("Loading properties from [{0}].", new Object[]{ file
            .getAbsolutePath() }));
        try {
          final Map properties = PropertyUtils.load(new FileInputStream(file));
          return properties;

        }
        catch (IOException ioex) {
          logger.error("Error loading properties.", ioex);
          return Collections.EMPTY_MAP;
        }
      }
      else {
        logger.warn(MessageFormat.format("[{0}] is not a file. Ignoring.", new Object[]{ file
            .getAbsolutePath() }));
        return Collections.EMPTY_MAP;
      }
    }
    else {
      return Collections.EMPTY_MAP;
    }
  }
  
  protected void processProperties(GeneratorContext context) {
    final ClassItem[] classes = context.getGrammar().getClasses();
    for (int index = 0; index < classes.length; index++) {
      final ClassItem classItem = classes[index];
      final ClassContext classContext = context.getClassContext(classItem);
      processProperties(classContext);
    }
  }

  protected void processProperties(final ClassContext classContext) {

    addClassEntry(classContext);

    final FieldUse[] allFieldUses = FieldUtils.getAllFieldUses(classContext);
    for (int index = 0; index < allFieldUses.length; index++) {
      final FieldUse fieldUse = allFieldUses[index];
      final FieldItem fieldItem = FieldUtils.getFieldItem(fieldUse);
      processProperties(classContext, fieldItem);
    }
  }

  protected void processProperties(final ClassContext classContext, final FieldItem fieldItem) {
    addFieldEntry(classContext, fieldItem);
    if (FieldUtils.isEnumField(classContext, fieldItem)) {
      addEnumEntry(classContext, fieldItem);
    }
  }
  
  public void writeProperties(GeneratorContext context) {
    final PackageContext[] packageContexts = getPackageContexts(context);

    for (int index = 0; index < packageContexts.length; index++) {
      final PackageContext packageContext = packageContexts[index];
      writeProperties(packageContext);
    }
  }

  public void writeProperties(final PackageContext packageContext) {

    final Locale[] locales = getLocales();
    for (int index = 0; index < locales.length; index++) {
      final Locale locale = locales[index];
      writeProperties(packageContext, locale);

    }
  }

  public void writeProperties(final PackageContext packageContext, final Locale locale) {
    logger.info("Generating properties for package ["
        + packageContext._package.name()
        + "] and locale ["
        + locale.toString()
        + "].");
    final Map properties = getProperties(packageContext, locale);
    final String draftPackageName = packageContext._package.name();
    final String packageName = draftPackageName == null ? "" : draftPackageName;
    
    final String packageDirectoryName = packageName.replace('.', File.separatorChar);
    
    final File packageDirectory = new File(exportDirectory, packageDirectoryName);
    
    packageDirectory.mkdirs();
    final String fileName = exportName + "_" + locale.toString() + ".properties";
    
    final File file = new File(packageDirectory, fileName);
    
    OutputStream os = null;
    
    try {
      os = new FileOutputStream(file);
      PropertyUtils.store(properties, os, null);
    }
    catch (IOException ex) {
      logger.error(MessageFormat.format("Error saving properties to [{0}].", new Object[]{ file
          .getAbsolutePath() }), ex);
    }
    finally {
      try {
        if (os != null) {
          os.close();
        }
      }
      catch (IOException ex) {
        // Ignore the exception
      }
    }
  }
  

  private Map properties = new HashMap();

  public Map getProperties(PackageContext packageContext, Locale locale) {
    final Map draftLocaleProperties = (Map) properties.get(locale);
    final Map localeProperties;
    if (draftLocaleProperties != null) {
      localeProperties = draftLocaleProperties;
    }
    else {
      localeProperties = new IdentityHashMap();
      properties.put(locale, localeProperties);
    }
    
    final Map draftPackageLocaleProperties = (Map) localeProperties.get(packageContext);
    final Map packageLocaleProperties;
    
    if (draftPackageLocaleProperties != null)
    {
      packageLocaleProperties = draftPackageLocaleProperties;
    }
    else
    {
      packageLocaleProperties = new TreeMap();
      localeProperties.put(packageContext, packageLocaleProperties);
    }
    
    return packageLocaleProperties;
  }
  
  public PackageContext[] getPackageContexts(final GeneratorContext generatorContext) {
    return generatorContext.getAllPackageContexts();
  }
  
}
