package org.jvnet.jaxbcommons.parent.addon;

import java.io.IOException;
import java.util.Collections;
import java.util.List;

import javax.xml.namespace.QName;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jvnet.jaxbcommons.util.AccessorUtils;
import org.jvnet.jaxbcommons.util.ClassUtils;
import org.jvnet.jaxbcommons.util.CustomizationUtils;
import org.jvnet.jaxbcommons.util.FieldUtils;
import org.jvnet.jaxbcommons.util.TypeUtils;
import org.jvnet.jaxbcommons.visitor.ClassifyingVisitor;
import org.xml.sax.ErrorHandler;

import com.sun.codemodel.JExpr;
import com.sun.codemodel.JFieldRef;
import com.sun.codemodel.JMethod;
import com.sun.codemodel.JOp;
import com.sun.tools.xjc.BadCommandLineException;
import com.sun.tools.xjc.CodeAugmenterEx;
import com.sun.tools.xjc.Options;
import com.sun.tools.xjc.generator.ClassContext;
import com.sun.tools.xjc.generator.GeneratorContext;
import com.sun.tools.xjc.grammar.AnnotatedGrammar;
import com.sun.tools.xjc.grammar.ClassItem;
import com.sun.tools.xjc.grammar.FieldItem;
import com.sun.tools.xjc.grammar.FieldUse;

public class AddOn implements CodeAugmenterEx {

	private static final QName PARENT_ELEMENT_NAME = new QName(
			"http://jaxbcommons.dev.java.net/parent", "parent");

	/**
	 * Logger.
	 */
	protected Log logger = LogFactory.getLog(AddOn.class);

	/**
	 * Does nothing.
	 * 
	 * @param opt
	 *            ignored.
	 * @param args
	 *            ignored.
	 * @param i
	 *            ignored.
	 * @return Alway returns <code>0</code>.
	 * @throws com.sun.tools.xjc.BadCommandLineException
	 *             Never thrown.
	 * @throws java.io.IOException
	 *             Never thrown.
	 */
	public int parseArgument(final Options opt, final String[] args, final int i)
			throws BadCommandLineException, IOException {
		return 0; // no option recognized
	}

	public String getOptionName() {
		return "Xparent";
	}

	public String getUsage() {
		return "  -Xparent:  ";
	}

	/**
	 * Augments the code.
	 * 
	 * @param grammar
	 *            annotated grammar.
	 * @param generatorContext
	 *            generator context.
	 * @param options
	 *            options.
	 * @param errorHandler
	 *            error handler.
	 * @return If the add-on executes successfully, return true. If it detects
	 *         some errors but those are reported and recovered gracefully,
	 *         return false.
	 */
	public boolean run(final AnnotatedGrammar grammar,
			final GeneratorContext generatorContext, final Options options,
			final ErrorHandler errorHandler) {

		logger.debug("Starting code augmentation.");

		// Process class items
		final ClassItem[] classItems = grammar.getClasses();
		for (int index = 0; index < classItems.length; index++) {
			final ClassItem classItem = classItems[index];

			final ClassContext classContext = generatorContext
					.getClassContext(classItem);
			logger.debug("Processing [" + classItem.name + "].");
			process(classContext);
		}
		logger.debug("Finished code augmentation.");
		return true;
	}

	protected void process(final ClassContext classContext1) {

		classContext1.target.visit(new ClassifyingVisitor(classContext1) {

			public Object onComplexSingleField(FieldItem fieldItem) {
				final ClassItem childClassItem = TypeUtils.getClassItem(
						classContext1, fieldItem);
				final ClassContext childClassContext = classContext1.parent
						.getClassContext(childClassItem);
				process(classContext1, fieldItem, childClassContext);
				return null;
			}

			public Object onComplexCollectionField(FieldItem fieldItem) {
				final ClassItem childClassItem = TypeUtils.getClassItem(
						classContext1, fieldItem);
				final ClassContext childClassContext = classContext1.parent
						.getClassContext(childClassItem);
				process(classContext1, fieldItem, childClassContext);
				return null;
			}

		});
	}

	protected void process(final ClassContext classContext,
			FieldItem fieldItem, ClassContext childClassContext) {
		final FieldItem childFieldItem = CustomizationUtils
				.getFieldItemWithCustomization(childClassContext,
						PARENT_ELEMENT_NAME);

		if (childFieldItem != null
				&& (classContext.target == TypeUtils.getClassItem(
						childClassContext, childFieldItem))) {
			process(classContext, fieldItem, childClassContext, childFieldItem);
		}
	}

	protected void process(ClassContext classContext, FieldItem fieldItem,
			ClassContext childClassContext, FieldItem childFieldItem) {
		logger.debug("Detected parent class ["
				+ ClassUtils.getClassName(classContext) + "] and field ["
				+ FieldUtils.getFieldName(classContext, fieldItem) + "]"
				+ ", child class ["
				+ ClassUtils.getClassName(childClassContext) + "] and field ["
				+ FieldUtils.getFieldName(childClassContext, childFieldItem)
				+ "].");
		
//		fieldItem.realization = new DefaultFieldRendererFactory(classContext.parent.getCodeModel())
//		{
//			public FieldRenderer create(ClassContext classContext, FieldUse fu) {
//		        if(fu.multiplicity.isAtMostOnce()) {
//		        	return new SingleFieldRenderer(classContext, fu);
//		        }
//		        else
//		        {
//				return super.create(classContext, fu);
//		        }
//			}
//		};
		
		final FieldUse fieldUse =FieldUtils.getFieldUse(classContext, fieldItem);
		final FieldUse childFieldUse =FieldUtils.getFieldUse(classContext, fieldItem);
		
		if (fieldUse.multiplicity.isAtMostOnce())
		{
			final JMethod setter = AccessorUtils.implSet(classContext, fieldItem);
			final JMethod childSetter = AccessorUtils.set(childClassContext, childFieldItem);
			final JFieldRef value = JExpr.ref("value");
			setter.body()._if(JOp.ne(value, JExpr._null()))._then().invoke(value, childSetter).arg(JExpr._this());
			
			final JMethod unsetter = AccessorUtils.implUnset(classContext, fieldItem);
			if (unsetter != null)
			{
				unsetter.body()._if(JOp.ne(value, JExpr._null()))._then().invoke(value, childSetter).arg(JExpr._null());
			}
		}
		else
		{
			
		}
	}


	public List getCustomizationURIs() {
		return Collections
				.singletonList("http://jaxbcommons.dev.java.net/parent");
	}
}
