package org.jvnet.jaxbcommons.tree.addon;

import java.io.IOException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jvnet.jaxbcommons.addon.generator.MethodStrategy;
import org.jvnet.jaxbcommons.tree.addon.generator.CreateNodeStrategy;
import org.xml.sax.ErrorHandler;

import com.sun.tools.xjc.BadCommandLineException;
import com.sun.tools.xjc.CodeAugmenter;
import com.sun.tools.xjc.Options;
import com.sun.tools.xjc.generator.ClassContext;
import com.sun.tools.xjc.generator.GeneratorContext;
import com.sun.tools.xjc.grammar.AnnotatedGrammar;
import com.sun.tools.xjc.grammar.ClassItem;

public class AddOn implements CodeAugmenter {

	protected Log logger = LogFactory.getLog(getClass());

	public int parseArgument(final Options opt, final String[] args, final int i)
			throws BadCommandLineException, IOException {
		return 0; // no option recognized
	}

	public String getOptionName() {
		return "Xtree";
	}

	public String getUsage() {
		return "  -Xtree";
	}

	/**
	 * Augments the code.
	 * 
	 * @param grammar
	 *            annotated grammar.
	 * @param generatorContext
	 *            generator context.
	 * @param options
	 *            options.
	 * @param errorHandler
	 *            error handler.
	 * @return If the add-on executes successfully, return true. If it detects
	 *         some errors but those are reported and recovered gracefully,
	 *         return false.
	 */
	public boolean run(final AnnotatedGrammar grammar,
			final GeneratorContext generatorContext, final Options options,
			final ErrorHandler errorHandler) {

		logger.debug("Starting code augmentation.");

		// Process class items
		final ClassItem[] classItems = grammar.getClasses();
		logger.debug("Processing [" + classItems.length + "] class items.");
		for (int index = 0; index < classItems.length; index++) {
			final ClassItem classItem = classItems[index];
			final ClassContext classContext = generatorContext
					.getClassContext(classItem);
			logger.debug("Processing [" + classItem.name + "].");
			getStrategy().generate(classContext, null, classContext.implClass);
		}
		logger.debug("Finished processing class items.");
		logger.debug("Finished code augmentation.");
		return true;
	}

	protected MethodStrategy strategy = new CreateNodeStrategy();

	public MethodStrategy getStrategy() {
		return strategy;
	}

	public void setStrategy(MethodStrategy strategy) {
		this.strategy = strategy;
	}
}
